#include <iostream>
using namespace std;

#include <iomanip>

#include <math.h>

#include "ddem_system.h"
#include "ddem_fsens_system.h"
#include "ddem_simulator.h"
#include "ddem_sensitivity_analyzer.h"
#include "ivp2dde_improved_CRK.h"
#include "ddem_sol_out.h"
#include "ddem_sens_out.h"
#include "ddem_error_codes.h"

#include "SYSTEMS/systems.h"

/* only one of these is necessary */
#include "SYSTEMS/system_01.h"
#include "SYSTEMS/system_02.h"
#include "SYSTEMS/system_03.h"
#include "SYSTEMS/system_04.h"
#include "SYSTEMS/system_05.h"
#include "SYSTEMS/system_06.h"
#include "SYSTEMS/system_07.h"
#include "SYSTEMS/system_08.h"

enum {DDEM_EXP_SIM, DDEM_EXP_SENS, DDEM_EXP_PARA};

#define round(x) ((int)floor((x) + 0.5))

int main(void)
{
	int experimentIdentifier = 1; /* used to select different experiments conveniently = 1,2,3,4*/
	
	
	ddemSystem* system;
  system = new ddemSystem();

	ddemSimulationAttributes* simA;
	

	double TOL = 1e-9;
	/* create the system and set all required functions and parameters*/
	switch(experimentIdentifier){
		case 1:
			getSystem_01(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_01(TOL, simA);
			break;
		case 2:
			getSystem_02(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_02(TOL, simA);
			break;
		case 3:
			getSystem_03(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_03(TOL, simA);
			break;
		case 4:
			getSystem_08(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_08(TOL, simA);
			break;
		default:
			cout << "Invalid experimentIdentifier " << endl;
			return -1;
	}

	int retCode;

	retCode = system->checkIntegrity();


	if(retCode < 0){
		cout << "The DDE system is not well defined." << endl <<
			      "Error Code = " << retCode << endl;
		return retCode;
	}else
		cout << "The DDE system is well defined." << endl;



	IVP2DDEImprovedCRK* improvedCRK;
	improvedCRK = new IVP2DDEImprovedCRK(0/*interp_Flag*/);

	ddemSimulator*	simulator;
	simulator = new ddemSimulator(improvedCRK);


	ddemSensitivitySolution* sensitivitySolution;
	sensitivitySolution = new ddemSensitivitySolution();
	
	ddemSensitivityAnalyzer* sensitivityAnalyzer;
	sensitivityAnalyzer = new ddemSensitivityAnalyzer();

	int* needSensitivity;
	needSensitivity = new int[system->getNParameters()];
	for(int l=0; l < system->getNParameters(); l++)
		needSensitivity[l] = DDEM_TRUE; 
	
	retCode =
		sensitivityAnalyzer->computeSensitivities(simulator,
									system,
									simA->tEnd,
									simA->parameters,
									simA->relTolerance,
									simA->absTolerance,
									simA->vanishingConst,
									needSensitivity,
									simA->comm,
									sensitivitySolution);
	
	cout << endl << endl << "Return code = " << retCode << endl;
	if(retCode != DDEM_SUCCESS)
		return retCode;

	ddemSensitivitySolutionFD* sensitivitySolutionFD;

	sensitivitySolutionFD = new ddemSensitivitySolutionFD();

	double deltaP = 1e-7;
	retCode =
			sensitivityAnalyzer->computeSensitivitiesFD(simulator,
									system,
									simA->tEnd,
									simA->parameters,
									deltaP,
									simA->relTolerance,
									simA->absTolerance,
									simA->vanishingConst,
									needSensitivity,
									1000,
									simA->comm,
									sensitivitySolutionFD);
	
	delete needSensitivity;

	cout << endl << endl << "Return code FD = " << retCode << endl;

	/*00000000000000000000000000000000000000000000000000000000000000000*/
	int outYIndex = 0;/* The chosen component for output = 0,1,..., (nVariables - 1) */
	int parameterIndex = 0;

	
	int interpSelect = DDEM_INTERP;/* DDEM_INTERP = value required, DDEM_DERIV_INTERP = derivative required*/

	string fileNameBase;
	string fileNameOutUniform;
	string fileNameOutUniformFD;
	
	fileNameBase = "OUT//sens_experiment_";
	fileNameBase.append(ddemInt2string(experimentIdentifier));
	fileNameBase.append("_");
	fileNameBase.append(ddemInt2string(round(-log10(TOL))));

	fileNameOutUniform = fileNameBase; 
	fileNameOutUniform.append("_U.dat");
	fileNameOutUniformFD = fileNameBase; 
	fileNameOutUniformFD.append("_");
	fileNameOutUniformFD.append(ddemInt2string(round(-log10(deltaP))));
	fileNameOutUniformFD.append("_U_FD.dat");

	/* hight precision uniform output : suitable for error analysis, ... */
	ddemOutSensUniform(sensitivitySolution,
											outYIndex,
											parameterIndex,
											interpSelect,
											1000/*nOutPoints*/,
											fileNameOutUniform,
											16  /*precision*/,
											simulator);

	ddemOutSensFDUniform(sensitivitySolutionFD,
												outYIndex,
												parameterIndex,
												fileNameOutUniformFD,
												16 /*precision*/);
	
	/* low precision non-uniform : suitable for plots*/
	string fileNameOutNonUniform;
	string fileNameOutNonUniformDisc;

	fileNameOutNonUniform = fileNameBase; 
	fileNameOutNonUniform.append("_NU.dat");

	fileNameOutNonUniformDisc = fileNameBase; 
	fileNameOutNonUniformDisc.append("_DISC.dat");

	ddemOutSensNonUniformAll(sensitivitySolution,
											interpSelect,
											1000/*nOutPoints*/,
											fileNameOutNonUniform,
											fileNameOutNonUniformDisc,
											8  /*precision*/,
											simulator);
	/*-------------- display integration statistics -----------*/
	cout << endl;
	cout << setiosflags( ios::scientific | ios::showpoint ) << setprecision(4); 
	cout << "TOL  = " << simA->relTolerance << endl;

	cout << endl;
	cout << "Number of successful steps = " << sensitivitySolution->statistics->nSteps << endl;
	cout << "Number of failed steps = " << sensitivitySolution->statistics->nFails << endl;
	cout << "Number of function(right hand side) evaluations = " << sensitivitySolution->statistics->nFEvals << endl;
	cout << "Number of vanishing steps = " << sensitivitySolution->statistics->nVanishingSteps << endl;
	cout << "Number of iteraions caused by vanishing steps = " << sensitivitySolution->statistics->nIteraions4VanishingSteps << endl;

	/* Display discontinuities along with their degree */
	cout << setiosflags( ios::scientific | ios::showpoint ) << setprecision(15); 
	cout << endl << "Tracked Discontinuities:" <<endl;
	for(int i=0; i < sensitivitySolution->nLambdas; i++)
		cout << "i = " << i << " lambda = " << 
		sensitivitySolution->meshPoints[sensitivitySolution->lambdasIndices[i]] <<
		" DiscDegree = " << sensitivitySolution->lambdasDiscDegrees[i] << endl;
	/*-------------------*/
	delete sensitivitySolution;
	delete sensitivitySolutionFD;
	delete simulator;
	delete sensitivityAnalyzer;
	delete improvedCRK;
	delete simA;
	delete system;
	/*-------------------*/
  return 0;
}
